@@ -1,4 +1,5 @@
 {-# LANGUAGE CPP #-}
+{-# LANGUAGE DeriveDataTypeable #-}
 ------------------------------------------------------------------------
 -- |
 -- Module      : Codec.Archive.Zip
@@ -34,6 +35,7 @@ module Codec.Archive.Zip
        , Entry (..)
        , CompressionMethod (..)
        , ZipOption (..)
+       , ZipException (..)
        , emptyArchive
 
        -- * Pure functions for working with zip archives
@@ -64,10 +66,13 @@ import Data.Binary
 import Data.Binary.Get
 import Data.Binary.Put
 import Data.List ( nub, find, intercalate )
+import Data.Data (Data)
+import Data.Typeable (Typeable)
 import Text.Printf
 import System.FilePath
 import System.Directory ( doesDirectoryExist, getDirectoryContents, createDirectoryIfMissing )
 import Control.Monad ( when, unless, zipWithM )
+import qualified Control.Exception as E
 import System.Directory ( getModificationTime )
 import System.IO ( stderr, hPutStrLn )
 import qualified Data.Digest.CRC32 as CRC32
@@ -76,10 +81,11 @@ import qualified Data.Map as M
 import Control.Applicative
 #endif
 #ifndef _WINDOWS
-import System.Posix.Files ( setFileTimes )
+import System.Posix.Files ( setFileTimes, setFileMode, fileMode, getFileStatus )
 #endif
 
 -- from bytestring
+import qualified Data.ByteString as S
 import qualified Data.ByteString.Lazy as B
 
 -- text
@@ -89,6 +95,13 @@ import qualified Data.Text.Lazy.Encoding as TL
 -- from zlib
 import qualified Codec.Compression.Zlib.Raw as Zlib
 
+versionMadeBy :: Word16
+#ifdef _WINDOWS
+versionMadeBy = 0x0000 -- FAT/VFAT/VFAT32 file attributes
+#else
+versionMadeBy = 0x0300 -- UNIX file attributes
+#endif
+
 #if !MIN_VERSION_binary(0, 6, 0)
 manySig :: Word32 -> Get a -> Get [a]
 manySig sig p = do
@@ -126,6 +139,7 @@ data Entry = Entry
                , eUncompressedSize        :: Word32              -- ^ Uncompressed size in bytes
                , eExtraField              :: B.ByteString        -- ^ Extra field - unused by this library
                , eFileComment             :: B.ByteString        -- ^ File comment - unused by this library
+               , eVersionMadeBy           :: Word16              -- ^ Version made by field
                , eInternalFileAttributes  :: Word16              -- ^ Internal file attributes - unused by this library
                , eExternalFileAttributes  :: Word32              -- ^ External file attributes (system-dependent)
                , eCompressedData          :: B.ByteString        -- ^ Compressed contents of file
@@ -143,6 +157,12 @@ data ZipOption = OptRecursive               -- ^ Recurse into directories when a
                | OptLocation FilePath Bool  -- ^ Where to place file when adding files and whether to append current path
                deriving (Read, Show, Eq)
 
+data ZipException =
+  CRC32Mismatch FilePath
+  deriving (Show, Typeable, Data)
+
+instance E.Exception ZipException
+
 -- | A zip archive with no contents.
 emptyArchive :: Archive
 emptyArchive = Archive
@@ -197,10 +217,7 @@ findEntryByPath path archive =
 -- | Returns uncompressed contents of zip entry.
 fromEntry :: Entry -> B.ByteString
 fromEntry entry =
-  let uncompressedData = decompressData (eCompressionMethod entry) (eCompressedData entry)
-  in  if eCRC32 entry == CRC32.crc32 uncompressedData
-         then uncompressedData
-         else error "CRC32 mismatch"
+  decompressData (eCompressionMethod entry) (eCompressedData entry)
 
 -- | Create an 'Entry' with specified file path, modification time, and contents.
 toEntry :: FilePath         -- ^ File path for entry
@@ -225,6 +242,7 @@ toEntry path modtime contents =
             , eUncompressedSize        = fromIntegral uncompressedSize
             , eExtraField              = B.empty
             , eFileComment             = B.empty
+            , eVersionMadeBy           = 0  -- FAT
             , eInternalFileAttributes  = 0  -- potentially non-text
             , eExternalFileAttributes  = 0  -- appropriate if from stdin
             , eCompressedData          = finalData
@@ -240,7 +258,7 @@ readEntry opts path = do
                                     _ | isDir -> "/"
                                       | otherwise -> "") in
               (case [(l,a) | OptLocation l a <- opts] of
-                    ((l,a):_) -> if a then l </> p else l
+                    ((l,a):_) -> if a then l </> p else l </> takeFileName p
                     _         -> p)
   contents <- if isDir
                  then return B.empty
@@ -252,16 +270,29 @@ readEntry opts path = do
   (TOD modEpochTime _) <- getModificationTime path
 #endif
   let entry = toEntry path' modEpochTime contents
+
+  entryE <-
+#ifdef _WINDOWS
+        return $ entry
+#else
+        do fm <- fmap fileMode $ getFileStatus path
+           let modes = fromIntegral $ shiftL (toInteger fm) 16
+           return $ entry { eExternalFileAttributes = modes,
+                            eVersionMadeBy = versionMadeBy }
+#endif
+
   when (OptVerbose `elem` opts) $ do
-    let compmethod = case eCompressionMethod entry of
+    let compmethod = case eCompressionMethod entryE of
                      Deflate       -> "deflated"
                      NoCompression -> "stored"
     hPutStrLn stderr $
-      printf "  adding: %s (%s %.f%%)" (eRelativePath entry)
-      compmethod (100 - (100 * compressionRatio entry))
-  return entry
+      printf "  adding: %s (%s %.f%%)" (eRelativePath entryE)
+      compmethod (100 - (100 * compressionRatio entryE))
+  return entryE
 
--- | Writes contents of an 'Entry' to a file.
+-- | Writes contents of an 'Entry' to a file.  Throws a
+-- 'CRC32Mismatch' exception if the CRC32 checksum for the entry
+-- does not match the uncompressed data.
 writeEntry :: [ZipOption] -> Entry -> IO ()
 writeEntry opts entry = do
   let path = case [d | OptDestination d <- opts] of
@@ -281,7 +312,16 @@ writeEntry opts entry = do
          hPutStrLn stderr $ case eCompressionMethod entry of
                                  Deflate       -> " inflating: " ++ path
                                  NoCompression -> "extracting: " ++ path
-       B.writeFile path (fromEntry entry)
+       let uncompressedData = fromEntry entry
+       if eCRC32 entry == CRC32.crc32 uncompressedData
+          then B.writeFile path uncompressedData
+          else E.throwIO $ CRC32Mismatch path
+#ifndef _WINDOWS
+       let modes = fromIntegral $ shiftR (eExternalFileAttributes entry) 16
+       when (eVersionMadeBy entry == 0x0300 &&
+             modes /= 0) $ setFileMode path modes
+#endif
+
   -- Note that last modified times are supported only for POSIX, not for
   -- Windows.
   setFileTimeStamp path (eLastModified entry)
@@ -465,7 +505,7 @@ getArchive = do
   locals <- manySig 0x04034b50 getLocalFile
   files <- manySig 0x02014b50 (getFileHeader (M.fromList locals))
   digSig <- lookAheadM getDigitalSignature
-#endif 
+#endif
   endSig <- getWord32le
   unless (endSig == 0x06054b50)
     $ fail "Did not find end of central directory signature"
@@ -566,20 +606,64 @@ getLocalFile = do
               skip 4 -- crc32
               cs <- getWord32le  -- compressed size
               skip 4 -- uncompressed size
-              if fromIntegral cs == length raw
-                 then return $ B.pack raw
-                 else fail "Content size mismatch in data descriptor record" 
+              if fromIntegral cs == B.length raw
+                 then return $ raw
+                 else fail "Content size mismatch in data descriptor record"
   return (fromIntegral offset, compressedData)
 
-getWordsTilSig :: Word32 -> Get [Word8]
-getWordsTilSig sig = go []
+getWordsTilSig :: Word32 -> Get B.ByteString
+#if MIN_VERSION_binary(0, 6, 0)
+getWordsTilSig sig = (B.fromChunks . reverse) `fmap` go Nothing []
+  where
+    sig' = S.pack [fromIntegral $ sig .&. 0xFF,
+                   fromIntegral $ sig `shiftR`  8 .&. 0xFF,
+                   fromIntegral $ sig `shiftR` 16 .&. 0xFF,
+                   fromIntegral $ sig `shiftR` 24 .&. 0xFF]
+    chunkSize = 16384
+    --chunkSize = 4 -- for testing prefix match
+    checkChunk chunk = do -- find in content
+          let (prefix, start) = S.breakSubstring sig' chunk
+          if S.null start
+            then return $ Right chunk
+            else return $ Left $ S.length prefix
+    go :: Maybe (Word8, Word8, Word8) -> [S.ByteString] -> Get [S.ByteString]
+    go prefixes acc = do
+      -- note: lookAheadE will rewind if the result is Left
+      eitherChunkOrIndex <- lookAheadE $ do
+          chunk <- getByteString chunkSize <|> B.toStrict `fmap` getRemainingLazyByteString
+          case prefixes of
+            Just (byte3,byte2,byte1) ->
+              let len = S.length chunk in
+                if len >= 1 &&
+                   S.pack [byte3,byte2,byte1,S.index chunk 0] == sig'
+                then return $ Left $ -3
+                else if len >= 2 &&
+                   S.pack [byte2,byte1,S.index chunk 0,S.index chunk 1] == sig'
+                then return $ Left $ -2
+                else if len >= 3 &&
+                   S.pack [byte1,S.index chunk 0,S.index chunk 1,S.index chunk 2] == sig'
+                then return $ Left $ -1
+                else checkChunk chunk
+            Nothing -> checkChunk chunk
+      case eitherChunkOrIndex of
+        Left index -> if index < 0
+            then do -- prefix match
+                skip (4 + index) -- skip over partial match in next chunk
+                return $ (S.take (S.length (head acc) + index) (head acc)) : (tail acc)
+            else do -- match inside this chunk
+                lastchunk <- getByteString index -- must read again
+                skip 4
+                return (lastchunk:acc)
+        Right chunk -> if len == chunkSize
+            then go prefixes' (chunk:acc)
+            else fail $ "getWordsTilSig: signature not found before EOF"
+          where
+            len = S.length chunk
+            prefixes' = Just $ (S.index chunk (len - 3), S.index chunk (len - 2), S.index chunk (len - 1))
+#else
+getWordsTilSig sig = B.pack `fmap` go []
   where
     go acc = do
-#if MIN_VERSION_binary(0, 6, 0)
-      (getWord32le >>= ensure (== sig) >> return (reverse acc)) <|>
-        do w <- getWord8
-           go (w:acc)
-#else
       sig' <- lookAhead getWord32le
       if sig == sig'
           then skip 4 >> return (reverse acc)
@@ -637,7 +721,7 @@ getFileHeader :: M.Map Word32 B.ByteString -- ^ map of (offset, content) pairs r
               -> Get Entry
 getFileHeader locals = do
   getWord32le >>= ensure (== 0x02014b50)
-  skip 2 -- version made by
+  vmb <- getWord16le  -- version made by
   versionNeededToExtract <- getWord8
   skip 1 -- upper byte indicates OS part of "version needed to extract"
   unless (versionNeededToExtract <= 20) $
@@ -678,6 +762,7 @@ getFileHeader locals = do
             , eUncompressedSize        = uncompressedSize
             , eExtraField              = extraField
             , eFileComment             = fileComment
+            , eVersionMadeBy           = vmb
             , eInternalFileAttributes  = internalFileAttributes
             , eExternalFileAttributes  = externalFileAttributes
             , eCompressedData          = compressedData
@@ -688,7 +773,7 @@ putFileHeader :: Word32        -- ^ offset
               -> Put
 putFileHeader offset local = do
   putWord32le 0x02014b50
-  putWord16le 0  -- version made by
+  putWord16le $ eVersionMadeBy local
   putWord16le 20 -- version needed to extract (>= 2.0)
   putWord16le 0x802  -- general purpose bit flag (bit 1 = max compression, bit 11 = UTF-8)
   putWord16le $ case eCompressionMethod local of
